#include<opencv2/opencv.hpp>
#include<iostream>
#include<cstdio>
#include<omp.h>
#include<math.h>
using namespace std;
using namespace cv;

#define T_ANGLE_THRE 10
#define T_SIZE_THRE 5

int vmin = 10, vmax = 256, smin = 30;
int trackObject = 0;
bool backprojMode = false;
bool selectObject = false;
bool showHist = true;
bool b_leftflag = false,b_rightflag=false;
Rect selection;
int g_left=0, g_right=0;

void brightAdjust(Mat &src, Mat &dst, double dContrast, double dBright)
{
	int nVal;
	int rowNumber = dst.rows;
	int colNumber = dst.cols*dst.channels();

	omp_set_num_threads(8);
#pragma omp parallel for

	for (int i = 0; i < rowNumber; i++)
	{
		uchar* dstdata = dst.ptr<uchar>(i);
		uchar* srcdata = src.ptr<uchar>(i);
		for (int j = 0; j < colNumber; j++)
		{
			nVal = ((dContrast * srcdata[j]) + dBright);
			if (nVal > 255) nVal = 255;
			else if (nVal < 0) nVal = 0;
			dstdata[j] = nVal;
		}
	}
}

void getDiffImage(Mat &src1, Mat &src2, Mat &dst, int nThre)
{
	int nVal;
	int rowNumber = src1.rows;
	int colNumber = src1.cols * src1.channels();

	omp_set_num_threads(8);
#pragma omp parallel for

	for (int i = 0; i < rowNumber; i++)
	{
		uchar* srcData1 = src1.ptr<uchar>(i);
		uchar* srcData2 = src2.ptr<uchar>(i);
		uchar* dstData = dst.ptr<uchar>(i);
		for (int j = 0; j < colNumber; j++)
		{
			if (srcData1[j] - srcData2[j]> nThre)
				dstData[j] = 255;
			else
				dstData[j] = 0;
		}
	}
}

vector<RotatedRect> armorDetect(vector<RotatedRect> vEllipse)
{
	vector<RotatedRect> vRlt;
	RotatedRect armor;
	int nL, nW;
	double dAngle;
	vRlt.clear();
	if (vEllipse.size() < 2)
		return vRlt;
	for (unsigned int nI = 0; nI < vEllipse.size() - 1; nI++)
	{
		for (unsigned int nJ = nI + 1; nJ < vEllipse.size(); nJ++)
		{
			dAngle = abs(vEllipse[nI].angle - vEllipse[nJ].angle);
			while (dAngle > 180)
				dAngle -= 180;
			if ((dAngle < T_ANGLE_THRE || 180 - dAngle < T_ANGLE_THRE) && abs(vEllipse[nI].size.height - vEllipse[nJ].size.height) < (vEllipse[nI].size.height + vEllipse[nJ].size.height) / T_SIZE_THRE && abs(vEllipse[nI].size.width - vEllipse[nJ].size.width) < (vEllipse[nI].size.width + vEllipse[nJ].size.width) / T_SIZE_THRE)
			{
				selection.x = min(vEllipse[nI].center.x - vEllipse[nI].size.width / 2, vEllipse[nJ].center.x - vEllipse[nJ].size.width / 2);
				selection.y = min(vEllipse[nI].center.y - vEllipse[nI].size.height / 2, vEllipse[nJ].center.y - vEllipse[nJ].size.height / 2);
				selection.height = fabs(vEllipse[nI].size.height - vEllipse[nJ].size.height) + (vEllipse[nI].size.height + vEllipse[nJ].size.height) / 2;
				selection.width = fabs(vEllipse[nI].center.x - vEllipse[nJ].center.x) + (vEllipse[nI].size.width + vEllipse[nJ].size.width) / 2;
				g_left= nI;
				g_right = nJ;
				b_leftflag = true;
				b_rightflag = true;
				armor.center.x = (vEllipse[nI].center.x + vEllipse[nJ].center.x) / 2;
				armor.center.y = (vEllipse[nI].center.y + vEllipse[nJ].center.y) / 2;
				armor.angle = (vEllipse[nI].angle + vEllipse[nJ].angle) / 2;
				if (180 - dAngle < T_ANGLE_THRE)
					armor.angle += 90;
				nL = (vEllipse[nI].size.height + vEllipse[nJ].size.height) / 2;
				nW = sqrt((vEllipse[nI].center.x - vEllipse[nJ].center.x) * (vEllipse[nI].center.x - vEllipse[nJ].center.x) + (vEllipse[nI].center.y - vEllipse[nJ].center.y) * (vEllipse[nI].center.y - vEllipse[nJ].center.y));
				if (nL < nW)
				{
					armor.size.height = nL;
					armor.size.width = nW;
				}
				else
				{
					armor.size.height = nW;
					armor.size.width = nL;
				}

				vRlt.push_back(armor);

				trackObject = -1;
			}
		}
	}
	return vRlt;
}

void drawBox(RotatedRect box, Mat &img)
{
	Point2f vertex[4];
	box.points(vertex);
	for (int i = 0; i < 4; i++)
	{
		line(img, vertex[i], vertex[(i + 1) % 4], Scalar(0, 0, 255), 2, CV_AA);
	}
}


int main()
{
	bool bFlag = true;
	int hsize = 16;
	float hranges[] = { 0,180 };
	const float* phranges = hranges;
	vector<Mat> channels;
	vector<RotatedRect> vEllipse;
	vector<Rect> rRect;
	vector<vector<Point> > contours;
	vector<Vec4i> hierarchy;
	vector<Vec2f> boxsize;
	vector<RotatedRect> vRlt;
	RotatedRect box;
	Rect rrect;

	Point2f vertex[4];
	Rect trackWindow_left, trackWindow_right;
	Mat frame, bImage, gImage, rImage, rawImage, grayImage, rlt;
	Mat binary;

	Mat hsv, hue, mask, hist, histimg = Mat::zeros(200, 320, CV_8UC3), backproj;
	namedWindow("Histogram", 0);

	VideoCapture cap("E:/RedCar.avi");
	cap >> frame;
	frame.copyTo(rawImage);
	bImage.copyTo(binary);
	double time = 0;
	unsigned int frames = 0;
	bool paused = false;
	while (1)
	{
		frames++;
		printf("%d\n", frames);
		double t0 = getTickCount();
		brightAdjust(frame, rawImage, 1, -120);

		split(rawImage, channels);
		bImage = channels.at(0);
		gImage = channels.at(1);
		rImage = channels.at(2);
		bImage.copyTo(binary);
		getDiffImage(rImage, gImage, binary, 25);
		Mat element = getStructuringElement(MORPH_RECT, Size(5, 5));
		dilate(binary, grayImage, element, Point(-1, -1), 3);
		erode(grayImage, rlt, element, Point(-1, -1), 1);
 		findContours(rlt, contours, hierarchy, CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE);
		if (contours.size() == 0)
			continue;
		for (int i = 0; i < contours.size(); i++)
		{
			if (contours.size()>10)
			//if ((contourArea(contours[i])>300 && contours.size()>10) || (contours.size()<10 && contours.size()>30))
			{

				box = minAreaRect(Mat(contours[i]));
				box.points(vertex);
				for (int i = 0; i < 4; i++)
					line(rlt, vertex[i], vertex[(i + 1) % 4], Scalar(255), 2, CV_AA);
				//box = cvFitEllipse2(&Mat(contours[i]));
				//ellipse(rlt,box,Scalar(255), 2, CV_AA);
				rrect =boundingRect(Mat(contours[i]));

				if ((box.size.height / box.size.width) > 1.0)
					bFlag = true;
				for (int nI = 0; nI < 5; nI++)
				{
					for (int nJ = 0; nJ < 5; nJ++)
					{
						if (box.center.y - 2 + nJ > 0 && box.center.y - 2 + nJ < 480 && box.center.x - 2 + nI > 0 && box.center.x - 2 + nI <  640)
						{
							Vec3b sx = frame.at<Vec3b>((int)(box.center.y - 2 + nJ), (int)(box.center.x - 2 + nI));
							if (sx[0] < 200 || sx[1] < 200 || sx[2] < 200)
							{
								int x1 = sx[0];
								int x2 = sx[1];
								int x3 = sx[2];
								bFlag = false;
								break;
							}
						}
					}
				}
				if (bFlag)
				{
					vEllipse.push_back(box);
					rRect.push_back(rrect);
				}
			}


		}
		vRlt = armorDetect(vEllipse);
		for (unsigned int nI = 0; nI < vRlt.size(); nI++)
			drawBox(vRlt[nI], frame);


		if (trackObject == -1)
		{
			rectangle(rawImage, selection, Scalar(255, 0, 0), 3, 8, 0);

		}
		cvtColor(rawImage, hsv, COLOR_BGR2HSV);


		//int _vmin = vmin, _vmax = vmax;

		//inRange(hsv, Scalar(0, smin, MIN(_vmin, _vmax)),
			//Scalar(180, 256, MAX(_vmin, _vmax)), mask);

		if (trackObject)
		{
			int ch[] = { 0, 0 };
			hue.create(hsv.size(), hsv.depth());
			mixChannels(&hsv, 1, &hue, 1, ch, 1);

			if (trackObject < 0)
			{
				Mat roi(hue, selection), maskroi(rlt, selection);
				calcHist(&roi, 1, 0, maskroi, hist, 1, &hsize, &phranges);
				normalize(hist, hist, 0, 255, CV_MINMAX);

				trackWindow_left = rRect[g_left];
				trackWindow_right= rRect[g_right];

				trackObject = 2;

				histimg = Scalar::all(0);
				int binW = histimg.cols / hsize;
				Mat buf(1, hsize, CV_8UC3);
				for (int i = 0; i < hsize; i++)
					buf.at<Vec3b>(i) = Vec3b(saturate_cast<uchar>(i*180. / hsize), 255, 255);
				cvtColor(buf, buf, CV_HSV2BGR);

				for (int i = 0; i < hsize; i++)
				{
					int val = saturate_cast<int>(hist.at<float>(i)*histimg.rows / 255);
					rectangle(histimg, Point(i*binW, histimg.rows),
						Point((i + 1)*binW, histimg.rows - val),
						Scalar(buf.at<Vec3b>(i)), -1, 8);
				}
			}
			calcBackProject(&hue, 1, 0, hist, backproj, &phranges);
			backproj &= rlt;
			if(b_leftflag)
			{
				RotatedRect trackBox_left = CamShift(backproj, trackWindow_left,
					TermCriteria(CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1));
				ellipse(rawImage, trackBox_left, Scalar(0, 0, 255), 3, CV_AA);
			}
				
			if (b_rightflag)
			{
				RotatedRect trackBox_right = CamShift(backproj, trackWindow_right,
					TermCriteria(CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1));
				ellipse(rawImage, trackBox_right, Scalar(0, 0, 255), 3, CV_AA);
			}

   			if (trackWindow_left.area() <= 1)
			{
				//int cols = backproj.cols, rows = backproj.rows, r = (MIN(cols, rows) + 5) / 6;
				//trackWindow_left = Rect(trackWindow_left.x - r, trackWindow_left.y - r,
				//	trackWindow_left.x + r, trackWindow_left.y + r) &
			    //Rect(0, 0, cols, rows);
				trackObject -= 1;
			    b_leftflag = 0;
			}
			if (trackWindow_right.area() <= 1)
			{
				//int cols = backproj.cols, rows = backproj.rows, r = (MIN(cols, rows) + 5) / 6;
				//trackWindow_left = Rect(trackWindow_right.x - r, trackWindow_right.y - r,
				//	trackWindow_right.x + r, trackWindow_right.y + r) &
				//	Rect(0, 0, cols, rows);
				trackObject -= 1;
				b_rightflag = 0;
			}

			if (backprojMode)
				cvtColor(backproj, rawImage, COLOR_GRAY2BGR);
		}
		if (trackObject < 0)
			paused = false;

		if (selectObject && selection.width > 0 && selection.height > 0)
		{
			Mat roi(rawImage, selection);
			bitwise_not(roi, roi);
		}

		imshow("rawImage", rawImage);
		imshow("Histogram", histimg);


		//imshow("B", bImage);
		//imshow("G", gImage);
		//imshow("R", rImage);
		//imshow("Bi", binary);
		//imshow("RLT", rlt);
		imshow("frame", frame);
		//imshow("1", frame);
		//imshow("2", rawImage);
		cvWaitKey(30);
		vEllipse.clear();
		vRlt.clear();
		rRect.clear();
		cap >> frame;
		time += (getTickCount() - t0) / getTickFrequency();
		cout << frames / time << " fps" << endl;


		char c = (char)waitKey(0);
 		if (c == 27)
			break;
		switch (c)
		{
		case 'b':
			backprojMode = !backprojMode;
			break;
		case 'c':
			trackObject = 0;
			histimg = Scalar::all(0);
			break;
		case 'h':
			showHist = !showHist;
			if (!showHist)
				destroyWindow("Histogram");
			else
				namedWindow("Histogram", 1);
			break;
		case 'p':
			paused = !paused;
			break;
		default:
			;
		}
	}
	return 0;
}